#TODO: apply_rw with gaps
function apply_rw(data::Array{Float64}, W::Matrix{Float64})
    nc = size(data, 1)
    W1 = W[1, 1:nc]
    W2 = W[2, 1:nc]

    data_r = data .* W1 .* W2
    return (data_r, W1 .* W2)
end

function apply_rw(data::Vector{<:Array{Float64}}, W::Vector{Matrix{Float64}})
    if length(W) != length(data)
        error("Lenghts must match")
    end
    nc = size.(data, 1)

    rw1 = [W[k][1, 1:nc[k]] for k=1:length(W)]
    rw2 = [W[k][2, 1:nc[k]] for k=1:length(W)]
    rw = [rw1[k] .* rw2[k] for k=1:length(W)]
    data_r = [data[k] .* rw[k] for k=1:length(data)]
    return (data_r, rw)
end

function check_corr_der(obs::Corr, derm::Vector{Corr})   
    g1 = Vector{String}(undef, 0)
    g2 = Vector{String}(undef, 0)

    for d in derm
        aux = [d.gamma[1], d.gamma[2]]
        push!(g1, aux[1][1:end-3])
        push!(g2, aux[2][1:end-3])
    end

    h = copy(derm)
    push!(h, obs)
    
    
    if any(getfield.(h, :y0) .!= getfield(h[1], :y0))
        return false
    end
    for s in [:kappa, :mu]
        for k = 1:2
            if any(getindex.(getfield.(h, s), k) .!= getindex(getfield(h[1], s), k))
                return false
            end
        end
    end
    #gamma check
    if any(g1 .!= obs.gamma[1]) || any(g2 .!= obs.gamma[2])
        return false
    end

    return true
end

@doc raw"""
    corr_obs(cdata::CData; real::Bool=true, rw::Union{Array{Float64, 2}, Nothing}=nothing, L::Int64=1, info::Bool=false)

    corr_obs(cdata::Array{CData, 1}; real::Bool=true, rw::Union{Array{Array{Float64, 2}, 1}, Nothing}=nothing, L::Int64=1, info::Bool=false)

Creates a `Corr` struct with the given `CData` struct `cdata` (`read_mesons`) for a single replica.
An array of `CData` can be passed as argument for multiple replicas.

The flag `real` select the real or imaginary part of the correlator.
If `rw` is specified, the method applies reweighting. `rw` is passed as a matrix of Float64 (`read_ms1`)
The correlator can be normalized with the volume factor if `L` is fixed.

The flag `info` provides extra output that contains information about the primary observables. The function returns the primary observables ``<WO>`` and ``<W>``
(it returns the observable <O> if rw=nothing)

```@example
#Single replica
data = read_mesons(path, "G5", "G5")
rw = read_ms1(path_rw)
corr_pp = corr_obs.(data)
corr_pp_r = corr_obs.(data, rw=rw)

#Single replica + Info
data = read_mesons(path, "G5", "G5")
rw = read_ms1(path_rw)
corr_pp, O = corr_obs(data[1], info=true)
corr_pp_r, WO, W = corr_obs(data[1], rw=rw, info=true)

#Two replicas
data = read_mesons([path_r1, path_r2], "G5", "G5")
rw1 = read_ms1(path_rw1)
rw2 = read_ms1(path_rw2)

corr_pp = corr_obs.(data)
corr_pp_r = corr_obs.(data, rw=[rw1, rw2])
```
"""
function corr_obs(cdata::CData; real::Bool=true, rw::Union{Array{Float64, 2}, Nothing}=nothing, L::Int64=1, info::Bool=false, idm::Union{Vector{Int64},Nothing}=nothing, nms::Int64=Int64(maximum(cdata.vcfg)))

    real ? data = cdata.re_data ./ L^3 : data = cdata.im_data ./ L^3
    nt = size(data)[2]

    nms_vec = cdata.vcfg
    delta_cnfg = nms_vec[2] - nms_vec[1]
    idm = isnothing(idm) ? collect(1:delta_cnfg:nms) : idm


    if isnothing(rw)
        # idm = isnothing(idm) ? collect(1:nms) : idm
        obs = [uwreal(data[:, x0], cdata.id, idm, nms) for x0 = 1:nt]
    else
        # idm = isnothing(idm) ? collect(1:nms) : idm
        data_r, W = apply_rw(data, rw)
        ow = [uwreal(data_r[:, x0], cdata.id, idm, nms) for x0 = 1:nt]
        W_obs = uwreal(W, cdata.id, idm, nms)
        obs = [ow[x0] / W_obs for x0 = 1:nt]
    end

    if info && !isnothing(rw)
        return (Corr(obs, cdata), ow, W_obs)
    elseif info && isnothing(rw)
        return (Corr(obs, cdata), obs)
    else
        return Corr(obs, cdata)
    end
end
#function corr_obs for R != 1
#TODO: vcfg with gaps
function corr_obs(cdata::Array{CData, 1}; real::Bool=true, rw::Union{Array{Array{Float64, 2}, 1}, Nothing}=nothing, L::Int64=1, info::Bool=false, idm::Union{Vector{Int64},Nothing}=nothing, nms::Union{Int64, Nothing}=nothing)
    nr = length(cdata)
    id = getfield.(cdata, :id)
    vcfg = getfield.(cdata, :vcfg)
    replica = Int64.(maximum.(vcfg))

    nms = isnothing(nms) ?  sum(replica) : nms
    if isnothing(idm)
        delta_vcfg = vcfg[1][2] - vcfg[1][1]
        idm = collect(1:delta_vcfg:replica[1]) 
        for ii in eachindex(cdata)[2:end]
            delta_vcfg = vcfg[ii][2] - vcfg[ii][1]
            aux = collect(1:delta_vcfg:replica[ii]) .+ replica[ii-1]
            append!(idm, aux)
        end
    end
    
    if !all(id .== id[1])
        error("IDs are not equal")
    end

    real ? data = getfield.(cdata, :re_data) ./ L^3 : data = getfield.(cdata, :im_data) ./ L^3

    nt = size(data[1])[2]

    if isnothing(rw)
        tmp = data[1]
        [tmp = cat(tmp, data[k], dims=1) for k = 2:nr]
        obs = [uwreal(tmp[:, x0], id[1], replica, idm, nms) for x0 = 1:nt]
    else
        data_r, W = apply_rw(data, rw)
        tmp = data_r[1]
        tmp_W = W[1]
        [tmp = cat(tmp, data_r[k], dims=1) for k = 2:nr]
        [tmp_W = cat(tmp_W, W[k], dims=1) for k = 2:nr]

        ow = [uwreal(tmp[:, x0], id[1], replica, idm, nms) for x0 = 1:nt]
        W_obs = uwreal(tmp_W, id[1], replica, idm, nms)
        obs = [ow[x0] / W_obs for x0 = 1:nt]
    end
    if info && !isnothing(rw)
        return (Corr(obs, cdata), ow, W_obs)
    elseif info && isnothing(rw)
        return (Corr(obs, cdata), obs)
    else
        return Corr(obs, cdata)
    end
end   

@doc raw"""
    corr_obs_TSM(cdata_sl::CData, cdata_corr::CData; real::Bool=true, rw::Union{Array{Float64, 2}, Nothing}=nothing, L::Int64=1, info::Bool=false, idm_sl::Union{Vector{Int64},Nothing}=nothing, idm_corr::Union{Vector{Int64},Nothing}=nothing, nms::Int64=Int64(maximum(cdata_sl.vcfg)))

    corr_obs(cdata::Array{CData, 1}; real::Bool=true, rw::Union{Array{Array{Float64, 2}, 1}, Nothing}=nothing, L::Int64=1, info::Bool=false)

Creates a `Corr` struct  for Truncated Solver Method (TSM) with the given `CData` structs `cdata_sl` and `cdata_ex` for a single replica.
Two arrays of `CData` can be passed as argument for multiple replicas.

The flag `real` select the real or imaginary part of the correlator.
If `rw` is specified, the method applies reweighting. `rw` is passed as a matrix of Float64 (`read_ms1`)
The correlator can be normalized with the volume factor if `L` is fixed.

The flag `info` provides extra output that contains information about the primary observables. The function returns the primary observables ``<WO>`` and ``<W>``
(it returns the observable <O> if rw=nothing)

The flags `idm_sl`, `idm_corr` and `nms` can be used if the observable is not measure in every configuration. In particular:
`idm_sl::Vector{Int64}` If given, label the configurations where the sloppy observables is measured. Not required if sloppy is full statistics.    
`idm_corr::Vector{Int64}` If given, label the configurations where the correction observables is measured. Not required if correction is full statistics.
`nms::Int64` Total number of measurements in the ensembles. By default, nms=maximum(cdata_sl.vcfg) it is assumed that the sloppy observable is full statistics, while the exact observable is not.
Typically, only `idm_corr` has to be passed, if the exact observable is not full statistics. 
By playing with these three flags you ensure that the Gamma method is still working even if the configurations between sloppy and correction do not match.


```@example
#Single replica
data_sloppy      = read_mesons(path_sl, "G5", "G5")
data_correction  = read_mesons_correction(path_corr, "G5", "G5")

rw = read_ms1(path_rw)
corr_pp = corr_obs_TSM.(data_sloppy, data_correction)
corr_pp_r = corr_obs_TSM.(data_sloppy, data_correction rw=rw)

#Single replica + Info
data_sloppy      = read_mesons(path_sl, "G5", "G5")
data_correction  = read_mesons_correction(path_corr, "G5", "G5")
rw = read_ms1(path_rw)
corr_pp, O = corr_obs_TSM(data_sloppy, data_correction, info=true)
corr_pp_r, WO, W = corr_obs(data_sloppy, data_correction,, rw=rw, info=true)

#Two replicas
data = read_mesons([path_r1, path_r2], "G5", "G5")
rw1 = read_ms1(path_rw1)
rw2 = read_ms1(path_rw2)

corr_pp = corr_obs.(data)
corr_pp_r = corr_obs.(data, rw=[rw1, rw2])
```
"""
function corr_obs_TSM(cdata_sl::CData, cdata_corr::CData; real::Bool=true, rw::Union{Array{Float64, 2}, Nothing}=nothing, L::Int64=1, info::Bool=false, idm_sl::Union{Vector{Int64},Nothing}=nothing, idm_corr::Union{Vector{Int64},Nothing}=nothing, nms::Int64=Int64(maximum(cdata_sl.vcfg)))
    # cdata1 is sloppy, cdata2 is correction 
    if cdata_sl.id != cdata_corr.id
        error("Error: cdata_sl id != cdata_corr id")
    end
    if cdata_sl.header != cdata_corr.header # Base.:(==) and Base.:(!=) are redifined in juobs_types.jl
        error("Error: cdata_sl header != cdata_corr header")
    end

    data1 = real ? cdata_sl.re_data   ./ L^3 : cdata_sl.im_data   ./ L^3
    data2 = real ? cdata_corr.re_data ./ L^3 : cdata_corr.im_data ./ L^3

    nt1 = size(data1, 2) # 1
    nt2 = size(data2, 2)
    id_sl   = cdata_sl.id
    id_corr = cdata_corr.id 
    # added automatic idsm detection - to be tested 
    nms_vec_sl   = cdata_sl.vcfg
    nms_vec_corr = cdata_corr.vcfg

    delta_cnfg_sl   = nms_vec_sl[2] - nms_vec_sl[1]
    delta_cnfg_corr = nms_vec_corr[2] - nms_vec_corr[1]

    idm_sl   = isnothing(idm_sl) ? collect(1:delta_cnfg_sl:nms_vec_sl[end]) : idm_sl
    idm_corr = isnothing(idm_corr) ? collect(1:delta_cnfg_corr:nms_vec_corr[end]) : idm_corr
    # end section to be tested 
    if isnothing(rw)
        idm_sl   = isnothing(idm_sl)   ? collect(1:nms) : idm_sl
        idm_corr = isnothing(idm_corr) ? collect(1:nms) : idm_corr
        obs1 = [uwreal(data1[:, x0], id_sl, idm_sl, nms) for x0 = 1:nt1]
        obs2 = [uwreal(data2[:, x0], id_corr, idm_corr, nms) for x0 = 1:nt2]

    else
        idm_sl   = isnothing(idm_sl)   ? collect(1:nms) : idm_sl
        idm_corr = isnothing(idm_corr) ? collect(1:nms) : idm_corr

        data1_r, W_sl = apply_rw(data1, rw)
        data2_r, W_corr = apply_rw(data2, rw)
        ow1 = [uwreal(data1_r[:, x0], id_sl, idm_sl, nms) for x0 = 1:nt1]
        ow2 = [uwreal(data2_r[:, x0], id_corr, idm_corr, nms) for x0 = 1:nt2]

        W_obs_sl   = uwreal(W_sl, id_sl, idm_sl, nms)
        W_obs_corr = uwreal(W_corr, id_corr, idm_corr, nms)

        obs1 = [ow1[x0] / W_obs_sl for x0 = 1:nt1]
        obs2 = [ow2[x0] / W_obs_corr for x0 = 1:nt2]
    end

    if info && !isnothing(rw)
        return (Corr(obs1 + obs2, cdata_sl), ow1, ow2, W_obs_sl)
    elseif info && isnothing(rw)
        return (Corr(obs1 + obs2, cdata_sl), obs1, obs2)
    else
        return Corr(obs1 + obs2, cdata_sl)
    end

end
function corr_obs_TSM(cdata1::Array{CData, 1}, cdata2::Array{CData, 1}; real::Bool=true, rw::Union{Array{Array{Float64, 2}, 1}, Nothing}=nothing, L::Int64=1, info::Bool=false, idm_sl::Union{Vector{Int64},Nothing}=nothing, idm_corr::Union{Vector{Int64},Nothing}=nothing, nms::Union{Int64, Nothing}=nothing)
    if any(getfield.(cdata1, :id) .!= getfield.(cdata2, :id))
        error("Error: cdata1 id != cdata2 id")
    end
    if any(getfield.(cdata1, :header) .!= getfield.(cdata2, :header)) # Base.:(==) and Base.:(!=) are redifined in juobs_types.jl
        error("Error: cdata1 header != cdata2 header")
    end
    
    id = getfield.(cdata1, :id)
    if !all(id .== id[1])
        error("IDs are not equal")
    end

    vcfg_sl      = getfield.(cdata1, :vcfg)
    replica_sl   = Int64.(maximum.(vcfg_sl))
    vcfg_corr    = getfield.(cdata2, :vcfg)
    replica_corr = Int64.(maximum.(vcfg_corr))
    
    nms = isnothing(nms) ? sum(replica_sl) : nms # assuming vcfg_sl >= vcfg_corr
    
    if isnothing(idm_sl)
        delta_vcfg = vcfg_sl[1][2] - vcfg_sl[1][1]
        idm_sl = collect(1:delta_vcfg:replica_sl[1])
        for ii in eachindex(cdata1)[2:end]
            delta_vcfg = vcfg_sl[ii][2] - vcfg_sl[ii][1]
            aux = collect(1:delta_vcfg:replica_sl[ii]) .+ replica_sl[ii-1]
            append!(idm_sl, aux)
        end
    end
    if isnothing(idm_corr)
        delta_vcfg = vcfg_corr[1][2] - vcfg_corr[1][1]
        idm_corr = collect(1:delta_vcfg:replica_corr[1])
        for ii in eachindex(cdata2)[2:end]
            delta_vcfg = vcfg_corr[ii][2] - vcfg_corr[ii][1]
            aux = collect(1:delta_vcfg:replica_corr[ii]) .+ replica_corr[ii-1]
            append!(idm_corr, aux)
        end
    end

    data1 = real ? getfield.(cdata1, :re_data) ./ L^3 : getfield.(cdata1, :im_data) ./ L^3
    data2 = real ? getfield.(cdata2, :re_data) ./ L^3 : getfield.(cdata2, :im_data) ./ L^3

    nt = size(data1[1], 2)

    if isnothing(rw)
        tmp1 = cat(data1..., dims=1)
        tmp2 = cat(data2..., dims=1)

        obs1 = [uwreal(tmp1[:, x0], id[1], replica_sl, idm_sl, nms) for x0 = 1:nt]
        obs2 = [uwreal(tmp2[:, x0], id[1], replica_corr, idm_corr, nms) for x0 = 1:nt]

    else
        data1_r, W = apply_rw(data1, rw)
        data2_r, W = apply_rw(data2, rw)

        tmp1 = cat(data1_r..., dims=1)
        tmp2 = cat(data2_r..., dims=1)
        tmp_W = cat(W..., dims=1)

        ow1 = [uwreal(tmp1[:, x0], id[1], replica_sl, idm_sl, nms) for x0 = 1:nt]
        ow2 = [uwreal(tmp2[:, x0], id[1], replica_corr, idm_corr, nms) for x0 = 1:nt]

        W_obs = uwreal(tmp_W, id[1], replica_sl, idm_sl, nms  )

        obs1 = [ow1[x0] / W_obs for x0 = 1:nt]
        obs2 = [ow2[x0] / W_obs for x0 = 1:nt]

    end
    if info && !isnothing(rw)
        return (Corr(obs1 + obs2, cdata1), ow1, ow2, W_obs)
    elseif info && isnothing(rw)
        return (Corr(obs1 + obs2, cdata1), obs1, obs2)
    else
        return Corr(obs1 + obs2, cdata1)
    end
end    
@doc raw"""
    corr_sym(corrL::Corr, corrR::Corr, parity::Int64=1)

Computes the symmetrized correlator using the left correlador `corrL` and the right correlator `corrR`. The source position
of `corrR` must be `T - 1 - y0`, where `y0` is the source position of `corrL`. 

```@example
pp_sym = corr_sym(ppL, ppR, +1)
a0p_sym = corr_sym(a0pL, a0pR, -1)
```
"""
function corr_sym(corrL::Corr, corrR::Corr, parity::Int64=1)
    T = length(corrL.obs)
    sym = [:kappa, :mu, :gamma]
    if corrL.y0 != T - 1 - corrR.y0
        error("Corr: Parameter mismatch")
    end
    for s in sym
        if getfield(corrL, s) != getfield(corrR, s)
            error("Corr: Parameter mismatch")
        end
    end
    if abs(parity) != 1
        error("incorrect value of parity (+- 1)")
    end

    res = (corrL.obs[1:end] + parity * corrR.obs[end:-1:1]) / 2
    return Corr(res, corrL.kappa, corrL.mu, corrL.gamma, corrL.y0)
end
#TODO: VECTORIZE, uwreal?
@doc raw"""
    md_sea(a::uwreal, md::Vector{Matrix{Float64}}, ow::uwreal, w::Union{uwreal, Nothing}=nothing, ws::ADerrors.wspace=ADerrors.wsg)

    md_sea(a::uwreal, md::Vector{Matrix{Float64}}, ow::Array{uwreal}, w::Union{uwreal, Nothing}=nothing, ws::ADerrors.wspace=ADerrors.wsg)

Computes the derivative of an observable A with respect to the sea quark masses.

``\frac{d <A>}{dm(sea)} = \sum_i \frac{\partial <A>}{\partial <O_i>}  \frac{d <O_i>}{d m(sea)}``


``\frac{d <O_i>}{dm(sea)} = <O_i> <\frac{\partial S}{\partial m}> - <O_i \frac{\partial S}{\partial m}>``

where ``O_i`` are primary observables. The function returns  ``\frac{\partial <A>}{\partial <O_i>}  \frac{d <O_i>}{d m(sea)}``, where
the primary observable ``O_i`` is specified as an input argument.

`ow` is a primary observable or a vector of primary observables that contains ``<OW>`` (``<O>`` if reweighting is not applied) and 
`w` is a primary obserable that contains the reweighting factors ``<W>`` (`w`=nothing if reweighting is not applied).

`md` is a vector that contains the derivative of the action ``S`` with respect
to the sea quark masses for each replica. `md[irep][irw, icfg]`

`md_sea` returns a tuple of uwreal observables ``(dA/dm_l, dA/dm_s)|_{sea}``, 
where ``m_l`` and ``m_s`` are the light and strange quark masses.

```@example
#Single replica
data = read_mesons(path, "G5", "G5")
md = read_md(path_md)
rw = read_ms1(path_rw)

corr_pp, wpp, w = corr_obs.(data[1], rw=rw, info=true)
m = meff(corr_pp[1], plat)
m_mdl, m_mds = md_sea(m, [md], wpp, w, ADerrors.wsg)
m_shifted = m + 2 * dml * m_mdl + dms * m_mds

#Two replicas
data = read_mesons([path_r1, path_r2], "G5", "G5")
md1 = read_md(path_md1)
md2 = read_md(path_md2)

corr_pp, pp = corr_obs(data[1], info=true)
m = meff(corr_pp[1], plat)
m_mdl, m_mds = md_sea(m, [md1, md2], pp, ADerrors.wsg)
m_shifted = m + 2 * dml * m_mdl + dms * m_mds
```
"""
function md_sea(a::uwreal, md::Vector{Matrix{Float64}}, ow::uwreal, w::Union{uwreal, Nothing}=nothing, ws::ADerrors.wspace=ADerrors.wsg)
    nid = neid(a)
    p = findall(t-> t==1, a.prop)

    if nid != 1
        @info("Error: neid > 1")
    end

    id = ws.map_nob[p]
    if !all(id .== id[1])
        @info("ids do not match")
    end
    id = ws.id2str[id[1]]
    
    ivrep = getfield.(ws.fluc[p], :ivrep)
    ivrep1 = fill(ivrep[1], length(ivrep))
    if !all(ivrep .== ivrep1)
        @info("ivreps do not match")
    end
    ivrep = ivrep[1]

    if length(md) != length(ivrep)
        @info("Nr obs != Nr md")
    end

    #md_aux as a Matrix + Automatic truncation
    md_aux = md[1][:, 1:ivrep[1]]
    for k = 2:length(md)
        md_aux = cat(md_aux, md[k][:, 1:ivrep[k]], dims=2)
    end
    
    nrw = size(md_aux, 1)
    uwerr(ow)
    ow_data = ow.mean .+ mchist(ow, id)

    if isnothing(w)
        der = derivative(a, ow)
        d1 = der != 0.0 ? der * (ow * uwreal(md_aux[1, :], id, ivrep) - uwreal(ow_data .* md_aux[1, :], id, ivrep)) : 0.0
        
        if nrw == 1
            return (d1, d1)
        elseif nrw == 2
            d2 = der != 0.0 ? der * (ow * uwreal(md_aux[2, :], id, ivrep) - uwreal(ow_data .* md_aux[2, :], id, ivrep)) : 0.0
            return (d1, d2)
        end
    else
        uwerr(w)
        der = derivative(a, ow) * w.mean
        w_data = w.mean .+ mchist(w, id)
        d1 = der != 0.0 ? der * (ow * uwreal(w_data .* md_aux[1, :], id, ivrep) / w^2 - uwreal(ow_data .* md_aux[1, :], id, ivrep) / w) : 0.0
        
        if nrw == 1
            return (d1, d1)
        elseif nrw == 2
            d2 = der != 0.0 ? der * (ow * uwreal(w_data .* md_aux[2, :], id, ivrep) / w^2 - uwreal(ow_data .* md_aux[2, :], id, ivrep) / w) : 0.0
            return (d1, d2)
        end
    end

    return nothing
end

function md_sea(a::uwreal, md::Vector{Matrix{Float64}}, ow::Array{uwreal}, w::Union{uwreal, Nothing}=nothing, ws::ADerrors.wspace=ADerrors.wsg)
    d = [md_sea(a, md, ow_, w, ws) for ow_ in ow]
    return (sum(getindex.(d, 1)), sum(getindex.(d, 2)))
end

@doc raw"""
    md_val(a::uwreal, obs::Corr, derm::Vector{Corr})

Computes the derivative of an observable A with respect to the valence quark masses.

``\frac{d <A>}{dm(val)} = \sum_i \frac{\partial <A>}{\partial <O_i>}  \frac{d <O_i>}{d m(val)}``

``\frac{d <O_i>}{dm(val)} = <\frac{\partial O_i}{\partial m(val)}>``

where ``O_i`` are primary observables 

`md` is a vector that contains the derivative of the action ``S`` with respect
to the sea quark masses for each replica. `md[irep][irw, icfg]`

`md_val` returns a tuple of `uwreal` observables ``(dA/dm_1, dA/dm_2)|_{val}``, 
where ``m_1`` and ``m_2`` are the correlator masses.

```@example
data = read_mesons(path, "G5", "G5", legacy=true)
data_d1 = read_mesons(path, "G5_d1", "G5_d1", legacy=true)
data_d2 = read_mesons(path, "G5_d2", "G5_d2", legacy=true)

rw = read_ms1(path_rw)

corr_pp = corr_obs.(data, rw=rw)
corr_pp_d1 = corr_obs.(data_d1, rw=rw)
corr_pp_d2 = corr_obs.(data_d2, rw=rw)
derm = [[corr_pp_d1[k], corr_pp_d2[k]] for k = 1:length(pp_d1)]

m = meff(corr_pp[1], plat)
m_md1, m_md2 = md_val(m, corr_pp[1], derm[1])
m_shifted = m + dm1 * m_md1 + dm2 * m_md2
```
"""
function md_val(a::uwreal, obs::Corr, derm::Vector{Corr})
    nid = neid(a)
    if nid != 1
        error("Error: neid > 1")
    end
    if length(derm) != 2
        error("Error: length derm != 2")
    end
    if !check_corr_der(obs, derm)
        error("Corr parameters does not match")
    end

    corr = getfield(obs, :obs)
    prop = getfield.(corr, :prop)

    if all(count.(prop) .== 1)
        der = [derivative(a, corr[k]) for k = 1:length(corr)]
    elseif all(count.(prop) .== 2)
        corr_der = getfield.(corr, :der)
        n = findall.(t-> t==1, prop)
        n = vcat(n'...)
        if all(n[:, 1] .== n[1, 1]) # find ow and w
            n_w = n[1, 1]
            n_ow = n[:, 2]
        else
            n_w = n[1, 2]
            n_ow = n[:, 1]
        end
        w_mean = 1 / getindex.(corr_der, n_ow)[1]
        #= Extra information (ow, w)
        ws = ADerrors.wsg
        ow_mean = - getindex.(corr_der, n_w) .* w_mean^2
        w_data = w_mean .+ ws.fluc[n_w].delta
        fluc_ow = getfield.(ws.fluc[n_ow], :delta)
        fluc_ow = vcat(fluc_ow'...)
        ow_data = ow_mean .+ ws.fluc[n_ow]
        =#
        der = length(a.der) >= maximum(n_ow) ? a.der[n_ow] * w_mean : 0.0
    else
        return nothing
    end
    derm1, derm2 = derm
    return (sum(der .* derm1.obs), sum(der .* derm2.obs))
end

function plat_av(obs::Vector{uwreal}, plat::Vector{Int64}, wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)
    isnothing(wpm) ? uwerr.(obs) : [uwerr(obs_aux, wpm) for obs_aux in obs]
    w = 1 ./ err.(obs)[plat[1]:plat[2]].^2
    av = sum(w .* obs[plat[1]:plat[2]]) / sum(w)
    return av 
end

@doc raw"""
    bayesian_av(fun::Function, y::Array{uwreal}, tmin_array::Array{Int64}, tmax_array::Array{Int64}, k::Int64, pl::Bool, data::Bool; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)

    bayesian_av(fun1::Function, fun2::Function, y::Array{uwreal}, tmin_array::Array{Int64}, tmax_array::Array{Int64}, k1::Int64, k2::Int64, pl::Bool, data::Bool; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)

    bayesian_av(fun::Array{Function}, y::Array{uwreal}, tmin_array::Array{Int64}, tmax_array::Array{Int64}, k::Array{Int64}, pl::Bool, data::Bool; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)

Computes bayesian average of data. For a given fit function, it explores choices of fit intervals, assigning each of them a weight. The function saves the `first` fit parameter of your function, and then it does the weighted average of it and assigns a systematic. See https://arxiv.org/abs/2008.01069 

The function takes as input the fit intervals to explore. 

`tmin_array` is an array of integers with the lower bounds on the fit intervals to explore, ***ordered from lower to higher***.

`tmax_array` is an array of integers with the upper bounds on the fit intervals to explore, ***ordered from lower to higher***.

`k` is the number of parameters of the fit function to use.

You can also use as input two fit functions, and two values of `k`, one for each function. Then, for each fit interval choice, the function explores the two fit functions. This means that for each fit interval choice you get two results: one for the first fit funcction, and another for the second. You can also use a vector of functions and a vector of k (numer of parameters of each funtion) to apply the bayesian averaging method to multiple functions.

The method returns two objects: first, the weighted average as an uwreal object, with mean value and statistichal error. The second object returned is the systematic error coming from the fit interval variation. If `data` is `true`, then returns 4 objects: weighted average, systematic error, a vector with the results of the fit for each fit interval choice, and a vector with the weights associated to each fit.

```@example
@.fun(x,p) = p[1] * x ^0
k = 1
tmin_array = [10,11,12,13,14,15]
tmax_array = [80,81,82,83,84,85]
(average, systematics, data, weights) = bayesian_av(fun,x,tmin_array,tmax_array,k,pl=true,data=true)

@.fun1(x,p) = p[1] * x ^0
@.fun2(x,p) = p[1] + p[2] * exp( - p[3] * (x))
k1 = 1
k2 = 3
tmin_array = [10,11,12,13,14,15]
tmax_array = [80,81,82,83,84,85]
(average, systematics) = bayesian_av(fun1,fun2,x,tmin_array,tmax_array,k1,k2)

@.fun1(x,p) = p[1] * x ^0
@.fun2(x,p) = p[1] + p[2] * exp( - p[3] * (x))
k1 = 1
k2 = 3
tmin_array = [10,11,12,13,14,15]
tmax_array = [80,81,82,83,84,85]
(average, systematics) = bayesian_av([fun1,fun2],x,tmin_array,tmax_array,[k1,k2])
```
"""
function bayesian_av(fun::Function, y::Array{uwreal}, tmin_array::Array{Int64}, tmax_array::Array{Int64}, k::Int64; pl::Bool=false, data::Bool=false,
    wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing, path_plt::Union{String,Nothing}=nothing, 
    plt_title::Union{Nothing,String}=nothing, label::Union{Nothing, LaTeXString}=nothing)
    
    weight_model = Array{Float64,1}()
    AIC = Array{Float64,1}()
    chi2chi2exp = Array{Float64,1}()
    p1 = Array{uwreal,1}()
    mods = Array{String,1}()
    Pval = Array{Float64,1}()

    if tmax_array[end] > length(y)
        error("Error: upper bound for the fits is bigger than last data point")
    end

    total = length(y)
    isnothing(wpm) ? uwerr.(y) : [uwerr(y[i],wpm) for i in 1:length(y)]
    
    for INDEX in tmin_array ## vary tmin
        for j in tmax_array ## vary tmax
             try
                x = [i for i in INDEX+1:1:j]	
	            yy = y[INDEX+1:1:j]
	            Ncut = total - length(x)
                dy = err.(yy)
                W = 1 ./ dy .^2

                p00 = [0.5 for i in 1:1:k]
                chisq = gen_chisq(fun,x,dy)
                fit = curve_fit(fun,x,value.(yy),W,p00)
                isnothing(wpm) ? (up,chi_exp) = fit_error(chisq,coef(fit),yy) : (up,chi_exp) = fit_error(chisq,coef(fit),yy,wpm)
                isnothing(wpm) ? uwerr(up[1]) : uwerr(up[1],wpm)
                chi2 = sum(fit.resid.^2) # * dof(fit) / chi_exp
                Q = pvalue(chisq, sum(fit.resid.^2), value.(up), yy, wpm=wpm, W=W, nmc=10000)
                # if chi2 / dof(fit) > 5.  || abs(value(up[1])) > 2.  || err(up[1])/value(up[1])*100 >= 10.
                    # error()
                # end
                # println(up[1], " ", up[2], " ", up[3])
                push!(Pval, Q)

                # push!(AIC, chi2 + 2*k + 2*Ncut)   # AIC criteria
                push!(AIC, chi2 - 2*chi_exp) # TIC criteria

                push!(chi2chi2exp, chi2 / dof(fit)) #substitute chi_exp with  dof(fit) for chi^2/dof
                push!(p1, up[1])
                push!(mods,string("[", INDEX+1, ",", j, "]"))
                # println("chi2_vs_exp " , chi2 / dof(fit),  " chi2/dof ", sum(fit.resid.^2)/dof(fit))
            catch e
                @warn string(":/ Negative window for error propagation at tmin = ", INDEX, ", tmax = ", j, "; skipping that point")
            end
        end
    end
    
    # compute all weights
    offset = minimum(AIC)
    AIC = AIC .- offset
    weight_model = exp.(-0.5 .* AIC)
    weight_model = weight_model ./ sum(weight_model)
    
    # sort weights and discard 5% tail 
    idxW = sortperm(weight_model, rev=true)
    cumulative_w = cumsum(weight_model[idxW])
    # println("acceptance in bayeasian_av set to 0.99, should be 0.95")
    idxcumw = findfirst(x->x>=0.95, cumulative_w)
    # println(length(idxW) - idxcumw)
    idxW = sort(idxW[1:idxcumw])
    
    # take only the accepted 
    AIC = AIC[idxW]
    Pval = Pval[idxW]
    chi2chi2exp  = chi2chi2exp[idxW]
    mods = mods[idxW]
    p1 = p1[idxW]
    weight_model = weight_model[idxW]
    weight_model ./= sum(weight_model)


    
    p1_mean = sum(p1 .* weight_model)
    isnothing(wpm) ? uwerr(p1_mean) : uwerr(p1_mean, wpm) 
    # println(sum(p1 .^ 2 .* weight_model))
    # println((sum(p1 .* weight_model)) ^ 2)
    systematic_err =  0.0 #sqrt(sum(p1 .^ 2 .* weight_model) - (sum(p1 .* weight_model)) ^ 2) 
    if pl   
        fig = figure(figsize=(10,7.5))

        subplots_adjust(hspace=0.1) 
        subplot(411)    
        if !isnothing(plt_title)
            title(plt_title)
        end
        ax1 = gca()                
        x = 1:length(p1)
        y = value.(p1)
        dy = err.(p1)
        v = value(p1_mean)
        e = err(p1_mean)

        fill_between(x, v-e, v+e, color="royalblue", alpha=0.2)
        errorbar(x, y, dy, fmt="^", mfc="none", color="navy", capsize=2)
        setp(ax1.get_xticklabels(),visible=false) # Disable x tick labels
        isnothing(label) ?  ylabel(L"$p_1$") : ylabel(label)
    
        subplot(412)
        ax2=gca()
        bar(x, weight_model, alpha=0.4, color="forestgreen", edgecolor="darkgreen", linewidth=1.5)
        setp(ax2.get_xticklabels(),visible=false) # Disable x tick labels
        # errorbar(mods, weight_model, 0*dy, color="green")
        ylabel(L"$W$")

        subplot(413)
        ax3=gca()
        setp(ax3.get_xticklabels(),visible=false) # Disable x tick labels

        bar(x, Pval, alpha=0.4, color="forestgreen", edgecolor="darkgreen", linewidth=1.5 )
        ylabel(L"$p\ value$")
        
        
        subplot(414)
        bar(x, chi2chi2exp, alpha=0.4, color="forestgreen", edgecolor="darkgreen", linewidth=1.5 )
        # ylabel(L"$\chi^2/\chi^2_{\mathrm{exp}}$")
        ylabel(L"$\chi^2/\mathrm{d.o.f.}$")
        xticks(x, mods, rotation=45)
        xlabel(L"$Models$")

        display(fig)
        if !isnothing(path_plt)
            #tt = plt_title *".pdf" 
            savefig(path_plt)
        end
        close()
    end

    if !data                   
        return (p1_mean, systematic_err)                                         
    else               
        FitStorage = Dict(
        "W"        => weight_model,
        "AIC"      => AIC .+ offset,
        "pval"     => Pval, 
        "chivsexp" => chi2chi2exp,
        "mods"     => mods
        )
        return (p1_mean, systematic_err, p1, FitStorage)    
    end 

end

function bayesian_av(fun1::Function, fun2::Function, y::Array{uwreal}, tmin_array::Array{Int64}, tmax_array::Array{Int64}, k1::Int64, k2::Int64; pl::Bool=false, data::Bool=false, wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)

    weight_model = Array{Float64,1}()
    AIC = Array{Float64,1}()
    chi2chi2exp = Array{Float64,1}()
    p1 = Array{uwreal,1}()
    mods = Array{String,1}()

    if tmax_array[end] > length(y)
        error("Error: upper bound for the fits is bigger than last data point")
    end

    total = length(y)
    isnothing(wpm) ? uwerr.(y) : for i in 1:length(y) uwerr(y[i],wpm) end
    
    for INDEX in tmin_array ## vary tmin
        for j in tmax_array ## vary tmax
            try
                x = [i for i in INDEX+1:1:j]	
	            yy = y[INDEX+1:1:j]
	            Ncut = total - length(x)
                dy = err.(yy)
                W = 1 ./ dy .^2

                p00 = [0.5 for i in 1:1:k1]
                chisq = gen_chisq(fun1,x,dy)
                fit = curve_fit(fun1,x,value.(yy),W,p00)
                isnothing(wpm) ? (up,chi_exp) = fit_error(chisq,coef(fit),yy) : (up,chi_exp) = fit_error(chisq,coef(fit),yy,wpm)
                isnothing(wpm) ? uwerr(up[1]) : uwerr(up[1],wpm)
                chi2 = sum(fit.resid.^2) * dof(fit) / chi_exp
                push!(AIC, chi2 + 2*k1 + 2*Ncut)
                push!(chi2chi2exp, chi2 / dof(fit))
                push!(p1, up[1])

                push!(mods,string("[", INDEX+1, ",", j, "]"))

                p00 = [0.5 for i in 1:1:k2]
                chisq = gen_chisq(fun2,x,dy)
                fit = curve_fit(fun2,x,value.(yy),W,p00)
                isnothing(wpm) ? (up,chi_exp) = fit_error(chisq,coef(fit),yy) : (up,chi_exp) = fit_error(chisq,coef(fit),yy,wpm)
                uwerr(up[1],wpm)
                chi2 = sum(fit.resid.^2) * dof(fit) / chi_exp
                push!(AIC, chi2 + 2*k2 + 2*Ncut)
                push!(chi2chi2exp, chi2 / dof(fit))
                push!(p1, up[1])
                println(up[end])
                push!(mods,string("[", INDEX+1, ",", j, "]"))

            catch e
                @warn string(":/ Negative window for error propagation at tmin = ", INDEX, ", tmax = ", j, "; skipping that point")
            end
        end
    end

    offset = minimum(AIC)
    AIC = AIC .- offset
    weight_model = exp.(-0.5 .* AIC)
    
    p1_mean = sum(p1 .* weight_model)/sum(weight_model) ; isnothing(wpm) ? uwerr(p1_mean) : uwerr(p1_mean,wpm) 
    weight_model = weight_model ./ sum(weight_model)
    systematic_err = sqrt(sum(p1 .^ 2 .* weight_model) - (sum(p1 .* weight_model)) ^ 2) 

    if pl                       
        x = 1:length(p1)
        y = value.(p1)
        dy = err.(p1)
        v = value(p1_mean)
        e = err(p1_mean)
        
        figure()
        fill_between(1:length(p1), v-e, v+e, color="green", alpha=0.75)
        errorbar(mods, y, dy, fmt="x", color="black")
        ylabel(L"$p_1$")
        xlabel(L"model")
    
        display(gcf())

        figure()
        errorbar(mods, weight_model, 0*dy, color="green")
        ylabel(L"$weight$")
        xlabel(L"model")
    
        display(gcf())
    end

    if !data                   
        return (p1_mean, systematic_err)                                         
    else               
        return (p1_mean, systematic_err, p1, weight_model)    
    end 

end

function bayesian_av(fun::Array{Function}, y::Array{uwreal}, tmin_array::Array{Int64}, tmax_array::Array{Int64}, k::Array{Int64}; pl::Bool=false,
    data::Bool=false, wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing, 
    path_plt::Union{String,Nothing}=nothing, plt_title::Union{Nothing,String}=nothing)

    weight_model = Array{Float64,1}()
    AIC          = Array{Float64,1}()
    chi2chi2exp  = Array{Float64,1}()
    p1           = Array{uwreal,1}()
    mods         = Array{String,1}()
    Pval         = Array{Float64,1}()

    if tmax_array[end] > length(y)
        error("Error: upper bound for the fits is bigger than last data point")
    end

    total = length(y)
    isnothing(wpm) ? uwerr.(y) : [uwerr(y[i],wpm) for i in 1:length(y)] 
    
    for INDEX in tmin_array ## vary tmin
        for j in tmax_array ## vary tmax
            try
                x = [i for i in INDEX+1:1:j]	
	            yy = y[INDEX+1:1:j]
	            Ncut = total - length(x)
                dy = err.(yy)
                W = 1 ./ dy .^2

                for indice in 1:length(fun)
                    p00 = [0.5 for i in 1:1:k[indice]]
                    chisq = gen_chisq(fun[indice],x,dy)
                    fit = curve_fit(fun[indice],x,value.(yy),W,p00)
                    isnothing(wpm) ? (up,chi_exp) = fit_error(chisq,coef(fit),yy) : (up,chi_exp) = fit_error(chisq,coef(fit),yy,wpm)
                    isnothing(wpm) ? uwerr(up[1]) : uwerr(up[1],wpm)
                    chi2 = sum(fit.resid.^2) * dof(fit) / chi_exp
                    Q = pvalue(chisq, sum(fit.resid.^2), value.(up), yy, wpm=wpm, W=W, nmc=10000)
                    if chi2 /  dof(fit) < 5.
                        # if err(up[1])/value(up[1])*100 >= 10.
                        # error()
                        # end
                        push!(Pval, Q)
                        push!(AIC, chi2 + 2*k[indice] + 2*Ncut)
                        push!(chi2chi2exp, chi2 / dof(fit))
                        push!(p1, up[1])
                        push!(mods,string("[", INDEX+1, ",", j, "]"))
                    end
                end

            catch e
                @warn string(":/ Negative window for error propagation at tmin = ", INDEX, ", tmax = ", j, "; skipping that point")
            end
        end
    end
    # compute all weights
    offset = minimum(AIC)
    AIC = AIC .- offset
    weight_model = exp.(-0.5 .* AIC)
    weight_model = weight_model ./ sum(weight_model)
    
    # sort weights and discard 5% tail 
    idxW = sortperm(weight_model, rev=true)
    cumulative_w = cumsum(weight_model[idxW])
    idxcumw = findfirst(x->x>=0.95, cumulative_w)
    println(length(idxW) - idxcumw)
    idxW = sort(idxW[1:idxcumw])
    
    # take only the accepted 
    AIC = AIC[idxW]
    Pval = Pval[idxW]
    chi2chi2exp  = chi2chi2exp[idxW]
    mods = mods[idxW]
    p1 = p1[idxW]
    weight_model = weight_model[idxW]
    weight_model /= sum(weight_model)
    
    p1_mean = sum(p1 .* weight_model) 
    isnothing(wpm) ? uwerr(p1_mean) : uwerr(p1_mean,wpm) 
    systematic_err = sqrt(sum(p1 .^ 2 .* weight_model) - (sum(p1 .* weight_model)) ^ 2) 

    if pl   
        subplots_adjust(hspace=0.1) 
        subplot(411)    
        if !isnothing(plt_title)
            title(plt_title)
        end
        ax1 = gca()                
        x = 1:length(p1)
        y = value.(p1)
        dy = err.(p1)
        v = value(p1_mean)
        e = err(p1_mean)

        fill_between(x, v-e, v+e, color="royalblue", alpha=0.2)
        errorbar(x, y, dy, fmt="^", mfc="none", color="navy", capsize=2)
        setp(ax1.get_xticklabels(),visible=false) # Disable x tick labels
        ylabel(L"$p_1$")
    
        subplot(412)
        ax2=gca()
        bar(x, weight_model, alpha=0.4, color="forestgreen", edgecolor="darkgreen", linewidth=1.5)
        setp(ax2.get_xticklabels(),visible=false) # Disable x tick labels
        # errorbar(mods, weight_model, 0*dy, color="green")
        ylabel(L"$W$")

        subplot(413)
        ax3=gca()
        setp(ax3.get_xticklabels(),visible=false) # Disable x tick labels

        bar(x, Pval, alpha=0.4, color="forestgreen", edgecolor="darkgreen", linewidth=1.5 )
        ylabel(L"$p\ value$")
        
        
        subplot(414)
        bar(x, chi2chi2exp, alpha=0.4, color="forestgreen", edgecolor="darkgreen", linewidth=1.5 )
        ylabel(L"$\chi^2/\chi^2_{\mathrm{exp}}$")
        xticks(x, mods, rotation=45)
        xlabel(L"$Models$")

        display(gcf())
        if !isnothing(path_plt)
            tt = plt_title *".pdf" 
            savefig(joinpath(path_plt, tt))
        end
        close()
    end
    

    if !data                   
        return (p1_mean, systematic_err)                                         
    else               
        FitStorage = Dict(
        "W"        => weight_model,
        "AIC"      => AIC .+ offset,
        "pval"     => Pval, 
        "chivsexp" => chi2chi2exp,
        "mods"     => mods
        )
        return (p1_mean, systematic_err, p1, FitStorage)
    end 

end

function lin_fit(x::Vector{<:Real}, v::Vector{Float64}, e::Vector{Float64})
    sig2 = e .* e
    S = sum(1 ./ sig2)
    Sx = sum(x ./ sig2)
    Sy = sum(v ./ sig2)
    Sxy = sum(v .* x ./ sig2)
    Sxx = sum(x .* x ./sig2)
    delta = S * Sxx - Sx*Sx
    par = [Sxx*Sy-Sx*Sxy, S*Sxy-Sx*Sy] ./delta
    #C = [[Sxx/delta, -Sx/delta], [-Sx/delta,  S/delta]]
    return par
end
@doc raw"""
    lin_fit(x::Vector{<:Real}, y::Vector{uwreal})

Computes a linear fit of uwreal data points y. This method return uwreal fit parameters and chisqexpected.

```@example
fitp, csqexp = lin_fit(phi2, m2)
m2_phys = fitp[1] + fitp[2] * phi2_phys
```
"""
function lin_fit(x::Vector{<:Real}, y::Vector{uwreal}; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)
    isnothing(wpm) ? uwerr.(y) : [uwerr(yaux, wpm) for yaux in y]
    par = lin_fit(x, value.(y), err.(y))
    chisq(p, d) = sum((d .- p[1] .- p[2].*x).^2 ./ err.(y) .^2)
    (fitp, csqexp) = fit_error(chisq, par, y)
    for i = 1:length(fitp)
        isnothing(wpm) ? uwerr(fitp[i]) : uwerr(fitp[i], wpm)
        print("\n Fit parameter: ", i, ": ")
        details(fitp[i])
    end
    println("Chisq / chiexp: ", chisq(par, y), " / ", csqexp, " (dof: ", length(x)-length(par),")")
    return (fitp, csqexp)
end

@doc raw"""
    x_lin_fit(par::Vector{uwreal}, y::Union{uwreal, Float64})

Computes the results of a linear interpolation/extrapolation in the x axis
"""
x_lin_fit(par::Vector{uwreal}, y::Union{uwreal, Float64}) = (y - par[1]) / par[2]
@doc raw"""
    y_lin_fit(par::Vector{uwreal}, y::Union{uwreal, Float64})

Computes the results of a linear interpolation/extrapolation in the y axis
"""
y_lin_fit(par::Vector{uwreal}, x::Union{uwreal, Float64}) = par[1] + par[2] * x

@doc raw"""
    fit_routine(model::Function, xdata::Array{<:Real}, ydata::Array{uwreal}, param::Int64=3; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)

    fit_routine(model::Function, xdata::Array{uwreal}, ydata::Array{uwreal}, param::Int64=3; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing, covar::Bool=false)

Given a model function with a number param of parameters and an array of `uwreal`,
this function fit ydata with the given `model` and print fit information
The method return an array `upar` with the best fit parameters with their errors.
The flag `wpm` is an optional array of Float64 of lenght 4. The first three paramenters specify the criteria to determine
the summation windows:

- `vp[1]`: The autocorrelation function is summed up to ``t = round(vp[1])``.

- `vp[2]`: The sumation window is determined using U. Wolff poposal with ``S_\tau = wpm[2]``

- `vp[3]`: The autocorrelation function ``\Gamma(t)`` is summed up a point where its error ``\delta\Gamma(t)`` is a factor `vp[3]` times larger than the signal.

An additional fourth parameter `vp[4]`, tells ADerrors to add a tail to the error with ``\tau_{exp} = wpm[4]``.
Negative values of `wpm[1:4]` are ignored and only one component of `wpm[1:3]` needs to be positive.
If the flag `covar`is set to true, `fit_routine` takes into account covariances between x and y for each data point.
```@example
@. model(x,p) = p[1] + p[2] * exp(-(p[3]-p[1])*x)
@. model2(x,p) = p[1] + p[2] * x[:, 1] + (p[3] + p[4] * x[:, 1]) * x[:, 2]
fit_routine(model, xdata, ydata, param=3)
fit_routine(model, xdata, ydata, param=3, covar=true)
```
"""
function fit_routine(model::Function, xdata::Array{<:Real}, ydata::Array{uwreal}, param::Int64=3; info::Bool=false, wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing,
    inv_cov::Union{Matrix{Float64}, Nothing}=nothing)

    isnothing(wpm) ? uwerr.(ydata) : [uwerr(yaux, wpm) for yaux in ydata]
    
    yval = value.(ydata)
    yer = err.(ydata)
    
    ## CODE HERE THE UPDATED VERSION OF CORRELATED FITS
    if isnothing(inv_cov)
        @info("Uncorrelated fit")
        chisq = gen_chisq(model, xdata, yer)
        fit = curve_fit(model, xdata, yval, 1.0 ./ yer.^2, fill(0.5, param))
        (upar, chi_exp) = isnothing(wpm) ? fit_error(chisq, coef(fit), ydata) : fit_error(chisq, coef(fit), ydata, wpm)
    else
        @info("Correlated fit")
        #covar_inv = inv(get_covariance(ydata)) 
        #covar_inv = (covar_inv + covar_inv')/2
        #covar_inv = inv(Symmetric(get_covariance(ydata)))
        chisq(par, dat) = gen_chisq_correlated(model, xdata, covar_inv, par, dat)
        fit = curve_fit(model, xdata, yval, covar_inv, fill(0.5, param))        
        #println(chisq(coef(fit), ydata))
        (upar, chi_exp) = isnothing(wpm) ? fit_error(chisq, coef(fit), ydata, W=covar_inv) : fit_error(chisq, coef(fit), ydata, wpm, W=covar_inv)
    end
    chi2_fit_res = sum(fit.resid.^2 )
    # compute and print single point contribution to chi2
    # println("\n")
    # for i in 1:length(fit.resid)
        # println((fit.resid[i])^2)
    # end
    # println("\n")

    println("chi2 from fit residual = ", chi2_fit_res)
    println("chi2 from chi2 function = ", chisq(coef(fit), ydata))


    #Info
    if info
        for i = 1:length(upar)
            isnothing(wpm) ? uwerr(upar[i]) : uwerr(upar[i], wpm)
            print("\n Fit parameter: ", i, ": ")
            details(upar[i])
        end
    end
    #chis2_corrected = (length(yval) - param) * chisq(coef(fit), ydata) / chi_exp
    chis2_corrected = (length(yval) - param) * chi2_fit_res / chi_exp

    #println("Chisq / chiexp: ", chisq(coef(fit), ydata), " / ", chi_exp, " (dof: ", length(yval) - param,")")
    println("Chisq / chiexp: ", chi2_fit_res, " / ", chi_exp, " (dof: ", length(yval) - param,")")
    println("Chisq corrected: ", chis2_corrected)
    println("Return: params, chi2, chi2exp")
    return upar, chi2_fit_res,  chi_exp
end

function fit_routine(model::Vector{Function}, xdata::Vector{Array{Float64, N}} where N, ydata::Vector{Array{uwreal, N}} where N, param::Int64; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing,
    correlated_fit::Bool=false)

    if !(length(model) == length(xdata) == length(ydata))
        error("Dimension mismatch")
    end
    N = length(model)

    dat = ydata[1]
    idx = Vector{Vector{Int64}}(undef, N)
    e = Vector{Vector{Float64}}(undef, N)
    j = 1
    for i = 1:N
        if isnothing(wpm)
            uwerr.(ydata[i])
        else
            # println(yaux for yaux in ydata[i])
            for yaux in ydata[i]
                [uwerr(yaux[k], wpm) for k in 1:length(yaux)]
            end
            # [[uwerr(yaux[k], wpm) for k in eachindex(yaux)] for yaux in ydata[i]]
        end
        
        e[i] = err.(ydata[i])
        if i > 1 
            dat = vcat(dat, ydata[i])
        end
        stp = j + length(ydata[i]) - 1
        idx[i] = collect(j:stp)
        j = stp + 1
    end
    if !correlated_fit
        chisq = (par, dat) -> sum([sum((dat[idx[i]] .- model[i](xdata[i], par)).^2 ./e[i].^2) for i=1:N])
    else
        inv_cov_tot = juobs.inv_covar(vcat(ydata...))  #[inv_covar_multi_id(dat[idx[ii]]) for ii in 1:N]
        function chisq(par, dat)
            res = Vector{uwreal}(undef,0)
            for ii in 1:N
                res =  vcat(res, dat[idx[ii]] .- model[ii](xdata[ii],par))
            end 
            chi = reshape(res, 1,:) * inv_cov_tot*res
            return chi[1]
        end
    end

    min_fun(t) = chisq(t, value.(dat))
    p = fill(0.5, param)
    #println(chisq(p,dat))
    sol = optimize(min_fun, p, LBFGS()) 
    #println(chisq(sol.minimizer,dat))
    
    if !correlated_fit
        @info("Uncorrelated fit ")
        (upar, chi2_exp) = isnothing(wpm) ? fit_error(chisq, sol.minimizer, dat) : fit_error(chisq, sol.minimizer, dat, wpm)
    else
        @info("Correlated fit ")

        #inv_cov_merge = zeros(size(inv_cov_tot[1],1)*N, size(inv_cov_tot[1],1)*N )
        #count = 1
        #for k in 1:N
        #    dim_tmp = size(inv_cov_tot[k], 1)
        #    inv_cov_merge[count:dim_tmp+count-1, count:dim_tmp+count-1] = inv_cov_tot[k]
        #    count +=dim_tmp
        #end
        (upar, chi2_exp) = isnothing(wpm) ? fit_error(chisq, sol.minimizer, dat, W=inv_cov_tot) : fit_error(chisq, sol.minimizer, dat, wpm, W=inv_cov_tot)
    end

    println("Chisq / chiexp: ", min_fun(sol.minimizer), " / ", chi2_exp, " (dof: ", length(dat) - param,")")
    chis2_corrected = (length(dat) - param) * min_fun(sol.minimizer) / chi2_exp
    println("Chisq corrected: ", chis2_corrected)
    
    @info("Returns: params, chi2, chi2_exp\n")
    return upar, min_fun(sol.minimizer)  ,  chi2_exp
end

function fit_routine(model::Function, xdata::Array{uwreal}, ydata::Array{uwreal}, param::Int64=3; 
    wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing, covar::Bool=false)
    
    Nalpha = size(xdata, 2) # number of x-variables
    Ndata = size(xdata, 1) # number of datapoints
    if isnothing(wpm)
        uwerr.(ydata)
        uwerr.(xdata)
    else
        [uwerr(yaux, wpm) for yaux in ydata] 
        [uwerr(xaux, wpm) for xaux in xdata]
    end

    yval = value.(ydata)
    yer = err.(ydata)
    xval = value.(xdata)
    xer = err.(xdata)

    dat = Vector{Float64}(undef, Ndata * (Nalpha+1))
    ddat = Vector{Float64}(undef, Ndata * (Nalpha+1))
    data = Vector{uwreal}(undef, Ndata * (Nalpha+1)) 

    for i = 1:Nalpha
        dat[(i-1)*Ndata+1:i*Ndata] = xval[:, i]
        ddat[(i-1)*Ndata+1:i*Ndata] = xer[:, i]
        data[(i-1)*Ndata+1:i*Ndata] = xdata[:, i]
    end
    dat[Nalpha*Ndata+1:end] = yval
    ddat[Nalpha*Ndata+1:end] = yer
    data[Nalpha*Ndata+1:end] = ydata

    # Guess
    fit = curve_fit(model, xval, yval, 1.0 ./ yer.^2, fill(0.5, param))

    # Generate chi2 + solver
    if covar
        aux = Vector{Vector{uwreal}}(undef, Ndata)
        for k = 1:Ndata
            aux[k] = Vector{uwreal}(undef, Nalpha+1)
            for i = 1:Nalpha
                aux[k][i] = xdata[k, i]
            end
            aux[k][Nalpha+1] = ydata[k]
        end

        C = isnothing(wpm) ? [ADerrors.cov(aux[k]) for k = 1:Ndata] : [ADerrors.cov(aux[k], wpm) for k = 1:Ndata]
        chisq_full_cov(p, d) = get_chi2_cov(model, d, C, p, Nalpha)
        min_fun_cov(t) = chisq_full_cov(t, dat)
        sol = optimize(min_fun_cov, vcat(fit.param, dat[1:Nalpha*Ndata]), method=LBFGS())
        
        (upar, chi2_exp) = isnothing(wpm) ? fit_error(chisq_full_cov, sol.minimizer, data) : fit_error(chisq_full_cov, sol.minimizer, data, wpm)
        println("Chisq / chiexp: ", min_fun_cov(sol.minimizer), " / ", chi2_exp, " (dof: ", length(ydata) - param,")")
        chis2_corrected = (length(ydata) - param) * min_fun_cov(sol.minimizer) / chi2_exp
        println("Chisq corrected: ", chis2_corrected)
        return upar, min_fun_cov(sol.minimizer) / chi2_exp

    else
        chisq_full(p, d) = get_chi2(model, d, ddat, p, Nalpha)
        min_fun(t) = chisq_full(t, dat)
        sol = optimize(min_fun, vcat(fit.param, dat[1:Nalpha*Ndata]), method=LBFGS())
        
        (upar, chi2_exp) = isnothing(wpm) ? fit_error(chisq_full, sol.minimizer, data) : fit_error(chisq_full, sol.minimizer, data, wpm)
        println("Chisq / chiexp: ", min_fun(sol.minimizer), " / ", chi2_exp, " (dof: ", length(ydata) - param,")")
        chis2_corrected = (length(ydata) - param) * min_fun(sol.minimizer) / chi2_exp
        println("Chisq corrected: ", chis2_corrected)
        return upar, min_fun(sol.minimizer) / chi2_exp
    end

    #### chisq_full, min_fun out of conditional ->
    #### COMPILER WARNING ** incremental compilation may be fatally broken for this module **

    # Info
    #for i = 1:length(upar)
    #    isnothing(wpm) ? uwerr(upar[i]) : uwerr(upar[i], wpm)
    #    print("\n Fit parameter: ", i, ": ")
    #    details(upar[i])
    #end
    
end

function global_fit_routine(models::Vector{Function}, xdata::Vector{Array{Float64, 1}} , ydata::Vector{Array{uwreal, 1}}, n_par::Int64; guess_param::Vector{Float64}=fill(0.5, n_par), log::Bool=true,
    wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing,
    inv_cov::Union{Matrix{Float64}, Nothing}=nothing)

    if !(length(models) == length(ydata))
        error("Dimension mismatch")
    end
    N_sets = length(models)             # number of datasets
    x_flat = xdata[1]                   # unrolling xdata 
    y_flat = ydata[1]                   # unrolling ydata

    chunk_idx = Vector{Vector{Int64}}(undef, N_sets)
    j = 1
    for i = 1:N_sets
        isnothing(wpm) ? uwerr.(ydata[i]) :  [uwerr(yaux, wpm) for yaux in ydata[i]]
        if i > 1 
            y_flat = vcat(y_flat, ydata[i])
            x_flat =vcat(x_flat, xdata[i])
        end
        stp = j + length(ydata[i]) - 1
        chunk_idx[i] = collect(j:stp)
        j = stp + 1
    end
    function flat_model(x_flat::Vector{Float64}, param::Vector{Float64})
        fx = similar(x_flat)
        for ii in 1:N_sets
            mod = models[ii]
            fx[chunk_idx[ii]] = mod(x_flat[chunk_idx[ii]], param)
        end
        return fx
    end
    function flat_chisq(param, data)
        if isnothing(inv_cov)
            chi = 0.0
            for ii in 1:N_sets
                mod = models[ii]
                chi += sum((data[chunk_idx[ii]] .- mod(x_flat[chunk_idx[ii]], param)).^2 ./ err.(y_flat[chunk_idx[ii]]).^2 )        
            end
            return chi
        else 
            # take pieces of covariance and build correlated chi2 piece by piec 
            # chi =0.0
            # for ii in 1:N_sets
                # mod = models[ii]
                # cov_inv = juovs.inv_covar(ydata[ii])
                # res = ydata[ii] .- mod(xdata[ii], param)
                # chi += (reshape(res, 1,:) * cov_inv * res)[1]
            # end
            # return chi
            # take whole covariance and multiply for whole vector of dat-mod(x,param)
            res = Vector{uwreal}(undef, 0)
            for ii in 1:N_sets
                mod = models[ii]
                res =vcat(res, data[chunk_idx[ii]] .- mod(x_flat[chunk_idx[ii]], param))
            end
            chi = reshape(res, 1,:) * inv_cov * res  
            return chi[1]
        end
    end

    if isnothing(inv_cov)
        @info("Uncorrelated fit")
        fit = curve_fit(flat_model, x_flat, value.(y_flat), 1.0 ./ err.(y_flat).^2, guess_param )
        (upar, chi2_exp) = isnothing(wpm) ? fit_error(flat_chisq, coef(fit) , y_flat) : fit_error(flat_chisq, coef(fit) , y_flat, wpm)
    else
        @info("Correlated fit")
        fit = curve_fit(flat_model, x_flat, value.(y_flat), inv_cov, guess_param )
        (upar, chi2_exp) = isnothing(wpm) ? fit_error(flat_chisq, coef(fit) , y_flat, W=inv_cov) : fit_error(flat_chisq, coef(fit) , y_flat, wpm, W=inv_cov)
    end
    println("chi2 from fit residue: ", sum(fit.resid.^2))
    chisq = flat_chisq(coef(fit), y_flat)  
    if log
        println("Converged param: ", fit.converged)
        println("Chisq / chiexp: ", chisq, " / ", chi2_exp, " (dof: ", dof(fit),")")
        chis2_corrected = (dof(fit)) * chisq / chi2_exp
        println("Chisq corrected: ", chis2_corrected)
    end
    return upar, value(chisq), chi2_exp, dof(fit)
end

function gen_chisq(f::Function, x::Array{<:Real}, err::Vector{Float64}) #constrained
    chisq(par, dat) = sum((dat .- f(x,par)).^2 ./err.^2)
    return chisq
end
function gen_chisq_correlated(f::Function, x::Array{<:Real}, covar_inv::Matrix, par, dat )
    chi2 = 0.0
    #delta = dat - f(x, par)
    #chi2 = (reshape(delta, 1,:) * covar_inv * delta )[1]
    for i in eachindex(dat)
        for j in eachindex(dat)
            #println(f(x,par)[i])
            #println(f(x,par)[j])
            #println(f(x[j,:],par))
            #println(f(x[i,:],par))
            chi2 = chi2 +  (dat[i] - f(x,par)[i]) * covar_inv[i,j] * (dat[j] - f(x,par)[j])
        end
    end
    #chisq(par, dat) = sum(
    #    transpose((dat .- f(x,par))) * covar_inv * (dat .- f(x,par))
    #) 
    return chi2
end

function get_chi2(f::Function, data, ddata, par, Nalpha) #full
    chi2 = 0.0

    Ndata = div(length(data), Nalpha+1)
    Npar = length(par) - Ndata * Nalpha
    p = par[1:Npar]

    for k = 1:Ndata
        xx = [par[Npar + k + (i-1)*Ndata] for i = 1:Nalpha]
        Cinv = zeros(Nalpha+1, Nalpha+1)
        [Cinv[i, i] = 1 / ddata[k + (i-1)*Ndata]^2 for i = 1:Nalpha+1]

        xx = [par[Npar + k + (i-1)*Ndata] for i = 1:Nalpha]
        delta = [data[k + (i-1)*Ndata] - xx[i] for i = 1:Nalpha]
        yy = f(xx', p)
        push!(delta, data[k + Nalpha*Ndata] - yy[1])

        chi2 += delta' * Cinv * delta
    end
    return chi2
end

function get_chi2_cov(f::Function, data, C, par, Nalpha) # full + cov
    chi2 = 0.0

    Ndata = div(length(data), Nalpha+1)
    Npar = length(par) - Ndata * Nalpha
    p = par[1:Npar]

    for k = 1:Ndata
        if det(C[k]) / prod(diag(C[k])) > 1e-6
            Cinv = inv(C[k])
        else
            Cinv = zeros(Nalpha+1, Nalpha+1)
            [Cinv[i, i] = 1 / C[k][i, i] for i = 1:Nalpha+1]
        end

        xx = [par[Npar + k + (i-1)*Ndata] for i = 1:Nalpha]
        delta = [data[k + (i-1)*Ndata] - xx[i] for i = 1:Nalpha]
        yy = f(xx', p)
        push!(delta, data[k + Nalpha*Ndata] - yy[1])
        
        chi2 += delta' * Cinv * delta 
    end
    return chi2
end

function get_covar(obs::Vector{uwreal}; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)
    isnothing(wpm) ? uwerr.(obs) :  [uwerr(yaux, wpm) for yaux in obs]
    ens = vcat(ensembles.(obs)...)
    tmp_fluc = mchist.(obs, ens)
    fluc = mapreduce(permutedims, vcat, tmp_fluc)' 
    covar = fluc' * fluc ./ size(fluc)[1]^2
    #@warn("should double check it. no justification to have a square in the Ncnfg")
    return covar
end
function get_covar_multi_id(obs::Vector{uwreal}; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)
    isnothing(wpm) ? uwerr.(obs) :  [uwerr(yaux, wpm) for yaux in obs]
    id_set = getindex.(ensembles.(obs), 2) # hardcoded assuming that ens label is always the second entry in each ensembles(obs[i])
    
    cov_tot = zeros(length(id_set), length(id_set))
    count = 1
    for (k,id) in enumerate(unique(id_set))
        idx       = findall(x-> x==id, id_set )
        fluc_aux  = mchist.(obs[idx], fill(id, length(idx)))
        fluc      = mapreduce(permutedims, vcat, fluc_aux)'
        covar_set = fluc' * fluc ./ size(fluc)[1]^2
        dim_tmp   = size(covar_set,1)
        cov_tot[count:dim_tmp+count-1, count:dim_tmp+count-1] = covar_set
        count +=dim_tmp
    end
    return cov_tot
end
function inv_covar(obs::Vector{uwreal}; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)
    cov = get_covar(obs, wpm=wpm)
    cov_inv = inv(cov)
    return (cov_inv' + cov_inv) / 2.
end
function inv_covar_multi_id(obs::Vector{uwreal}; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)
    cov = get_covar_multi_id(obs, wpm=wpm)
    cov_inv = inv(cov)
    return (cov_inv' + cov_inv) / 2.
end

@doc raw"""
    pvalue(chisq::Function,
    chi2::Float64,
    xp::Vector{Float64}, 
    data::Vector{uwreal},
    wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing} = Dict{Int64,Vector{Float64}}();
    W::Union{Vector{Float64},Array{Float64,2}} = Vector{Float64}(),
    nmc::Int64 = 5000)

Computes the p-value of a previously done fit, using as input the `\chi^2` observed from the fit, the fit parameters and the fitted data. 
The p-value for a given `\chi^2` is the probability of, given the data you have, finding such a `\chi^2` or worse from a fit, and still
have the data well described by the fit function. `nmc` is the number of MC samples used to estimate the p-value integral, default is 5000.
By now it only works with a vector for weights (containing the diagonal of W)

```@example
function fit_defs(f::Function,x,W)
    chisq(p,d)=sum((d-f(x,p)).^2 .*W)
    return chisq
end

@.fun(x,p) = p[1] + p[2] * x
chisq = fit_defs(fun, x, 1.0 ./ err.(y) .^ 2)
fit = curve_fit(fun, x, value.(y), 1.0 ./ err.(y) .^ 2, [0.5, 0.5])
(up, chiexp) = fit_error(chisq, coef(fit), y)

wpm = Dict{Int64,Vector{Float64}}()
wpm[1] = [-1.0,-1.0,4-0,-1.0]
Q = pvalue(chisq, chi2, value.(up), y, wpm; W = 1.0 ./ err.(y) .^ 2, nmc=10000)
#Q = pvalue(chisq, chi2, value.(up), y; W = 1.0 ./ err.(y) .^ 2, nmc=10000)
#Q = pvalue(chisq, chi2, value.(up), y)
```
"""
function pvalue(chisq::Function,
    chi2::Float64,
    xp::Vector{Float64}, 
    data::Vector{uwreal};
    wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing} =  Dict{Int64,Vector{Float64}}(),
    W::Union{Vector{Float64},Array{Float64,2}} = Vector{Float64}(),
    nmc::Int64 = 5000)

    n = length(xp)   # Number of fit parameters
    m = length(data) # Number of data

    xav = zeros(Float64, n+m)
    for i in 1:n
        xav[i] = xp[i]
    end
    for i in n+1:n+m
        xav[i] = data[i-n].mean
    end
    ccsq(x::Vector) = chisq(view(x, 1:n), view(x, n+1:n+m)) 
    if (n+m < 4)
        cfg = ForwardDiff.HessianConfig(ccsq, xav, ADerrors.Chunk{1}());
    else
        cfg = ForwardDiff.HessianConfig(ccsq, xav, ADerrors.Chunk{4}());
    end
        
    hess = Array{Float64}(undef, n+m, n+m)
    ForwardDiff.hessian!(hess, ccsq, xav, cfg)
        
    cse = 0.0
    Q = dQ = 0.0
    if (m-n > 0)
        if (length(W) == 0)
            Ww = zeros(Float64, m)
            for i in 1:m
                if (data[i].err == 0.0)
                    #isnothing(wpm) ? wuerr(data[i]) : uwerr(data[i], wpm)
                    uwerr(data[i], wpm)
                    if (data[i].err == 0.0)
                        error("Zero error in fit data")
                    end
                end
                Ww[i] = 1.0 / data[i].err^2
            end
        else
            Ww = W
        end
        #cse = chiexp(hess, data, Ww, wpm)

        m = length(data)
        n = size(hess, 1) - m

        hm = view(hess, 1:n, n+1:n+m)
        sm = Array{Float64, 2}(undef, n, m)

        for i in 1:n, j in 1:m
            sm[i,j] = hm[i,j] / sqrt.(Ww[j])
        end
        maux = sm * sm'
        hi   = LinearAlgebra.pinv(maux)
        Px   = -hm' * hi * hm

        for i in 1:m
            Px[i,i] = Ww[i] + Px[i,i]
        end
        C = cov(data) 
        
        nu = sqrt(C) * Px * sqrt(C)
        
        N = length(nu[1,:])
        z = randn(N, nmc)

        eig = abs.(eigvals(nu))
        eps = 1e-14 * maximum(eig)
        eig = eig .* (eig .> eps)

        aux = eig' * (z .^ 2)
        Q = 1.0 - juobs.mean(aux .< chi2)

        x = chi2 .- eig[2:end]' * (z[2:end,:].^2)
        x = x / eig[1]
        #dQ = juobs.mean((x .> 0) .* exp.(-x * 0.5) * 0.5 ./ sqrt.(abs.(x)))
        #dQ = err(cse)/value(cse) * dQ
    end

    return Q #uwreal([Q,dQ],"")
end